
Require Import ZArith.
Require Import List.
Require Import Recdef. (*Function*)

(* ============================================================== *)

Locate "_ * _".
Locate "_ + _".

Print Scope nat_scope.
Print Scope Z_scope.
Print Scope list_scope.

Check 3.
Eval compute in 4+5.
Check (3*5)%Z.
Eval compute in (3 + 5)%Z.

Open Scope Z_scope.

Check 3.
Check (S (S (S O))).
Eval compute in 7*3.

Close Scope Z_scope.

Check 3.


Search le.
SearchAbout le.
SearchPattern (le (_ + _) (_ + _)).
SearchPattern (_ + _ <= _ + _).
SearchRewrite (_ + (_ - _)).

(* ============================================================== *)

Definition comp : forall A B C:Set, (A->B) -> (B->C) -> A -> C
  := fun A B C f g x => g (f x).

Definition example (A:Set) (f:nat->A) := comp _ _ _ S f.

Set Implicit Arguments.

Definition comp1 : forall A B C:Set, (A->B) -> (B->C) -> A -> C
  := fun A B C f g x => g (f x).

Definition example1 (A:Set) (f:nat->A) := comp1 S f.

Check (comp1 (C:=nat) S).

Check (@comp1 nat nat nat S S).

Unset Implicit Arguments.

Definition comp2 : forall A B C:Set, (A->B) -> (B->C) -> A -> C
  := fun A B C f g x => g (f x).

Implicit Arguments comp2 [A C].

Definition example2 (A:Set) (f:nat->A) := comp2 nat S f.

Print Implicit example2.
Print Implicit comp2.


Set Implicit Arguments.


(* ============================================================== *)

(* Exercise: *)

Inductive InL (A:Type) (y:A) : list A -> Prop :=
| InHead : forall xs:list A, InL y (cons y xs)
| InTail : forall (x:A) (xs:list A), InL y xs -> InL y (cons x xs).


Fixpoint elem (a:Z) (l:list Z) {struct l} : bool :=   (* !!!!!!!!!!! *)
    match l with
      | nil => false
      | cons x xs => if Z_eq_dec x a then true else (elem a xs)
    end.

(* Prove the following theorem: *)

Theorem elem_corr : forall (a:Z) (l1 l2:list Z), elem a (app l1 l2) = orb (elem a l1) (elem a l2).
induction l1.
simpl. reflexivity.
intros. simpl.
elim (Z_eq_dec a0 a).
SearchAbout orb.
(*simpl.*)
rewrite Bool.orb_true_l.
trivial.
auto.
Qed.




(* ============================================================== *)

Definition head (A:Type) (l:list A) (p:l<>nil) : A.
refine (fun A l p=>
  match l return (l<>nil->A) with
  | nil => fun H => _
  | cons x xs => fun H => x
  end p).
elim H; reflexivity.
Defined.

Print Implicit head.

Definition headPre (A:Type) (l:list A) : Prop := l<>nil.

Inductive headRel (A:Type) (x:A) : list A -> Prop :=
  headIntro : forall l, headRel x (cons x l).

Print Implicit headRel.

Theorem head_correct : forall (A:Type) (l:list A) (p:headPre l), headRel (head  p) l.
Proof.
induction l.
intro H; elim H; reflexivity.
intros; destruct l; [simpl; constructor | simpl; constructor].
Qed.


(* EXTRACTION *)

Check head.

Extraction Language Haskell.
Extraction Inline False_rect.
Extraction head.


(* ============================================================== *)

(* An inductive relation "x is the last element of list l" *)
Inductive Last (A:Type) (x:A) : list A -> Prop :=
| last_base : Last x (cons x nil)
| last_step : forall l y, Last x l -> Last x (cons y l).


(* EXTRACTION *)

Theorem last_correct : forall (A:Type) (l:list A), l<>nil -> { x:A | Last x l }.
induction l.
intro H. elim H. reflexivity.
intros. destruct l.
exists a; auto.
constructor.
elim IHl.
intros. exists x.
constructor. assumption.
discriminate.
Qed.

Extraction Language Haskell.

Extraction Inline False_rect.
Extraction Inline sig_rect.
Extraction Inline list_rect.

Extraction last_correct.

(* ============================================================== *)

(* INSERTION SORT *)

Open Scope Z_scope. (* se no <=  interpretado em nat *)

Inductive Sorted : list Z -> Prop := 
  | sorted0 : Sorted nil 
  | sorted1 : forall z:Z, Sorted (z :: nil) 
  | sorted2 : 
      forall (z1 z2:Z) (l:list Z), 
        z1 <= z2 -> 
        Sorted (z2 :: l) -> Sorted (z1 :: z2 :: l). 


Fixpoint count (z:Z) (l:list Z) {struct l} : nat :=
  match l with
  | nil => 0%nat
  | (z' :: l') =>
      match Z_eq_dec z z' with
      | left _ => S (count z l')
      | right _ => count z l'
      end
  end.


Definition Perm (l1 l2:list Z) : Prop :=
  forall z, count z l1 = count z l2.

(*
 Exercise: prove that Perm is an equivalence relation (i.e. is reflexive, symmetric and transitive)

Lemma Perm_reflex : forall l:list Z, Perm l l.
Lemma Perm_sym : forall l1 l2, Perm l1 l2 -> Perm l2 l1.
Lemma Perm_trans : forall l1 l2 l3, Perm l1 l2 -> Perm l2 l3 -> Perm l1 l3.

*)




(*
 Exercise: prove the following lemmas:

Lemma Perm_cons : forall a l1 l2, Perm l1 l2 -> Perm (a::l1) (a::l2).
Lemma Perm_cons_cons : forall x y l, Perm (x::y::l) (y::x::l).

*)


Fixpoint insert (x:Z) (l:list Z) {struct l} : list Z :=
  match l with
    nil => cons x nil
  | cons h t =>
        match Z_lt_ge_dec x h with
          left _ => cons x (cons h t)
        | right _ => cons h (insert x t)
        end
  end.


Fixpoint isort (l:list Z) : list Z :=
  match l with
    nil => nil
  | cons h t => insert h (isort t)
  end.


(*
Theorem isort_correct : forall (l l':list Z),
  l'=isort l -> Perm l l' /\ Sorted l'.
Proof.
induction l; intros.
unfold Perm; rewrite H; split; auto.
simpl.
constructor.
simpl in H.
rewrite H. (* ??????????? *)
*)


Lemma insert_Perm : forall x l, Perm (x::l) (insert x l).
Proof.
unfold Perm; induction l.
simpl. reflexivity.
simpl insert. elim (Z_lt_ge_dec x a).  reflexivity. 
intros. rewrite Perm_cons_cons.
pattern (x::l). simpl count. elim (Z_eq_dec z a). 
intros.
rewrite IHl; reflexivity.
intros. apply IHl.
Qed.


Lemma insert_Sorted : forall x l, Sorted l -> Sorted (insert x l).
Proof.
intros x l H; elim H; simpl. 
constructor.

intro z; elim (Z_lt_ge_dec x z); intros.
constructor.
auto with zarith.
constructor.

constructor.
auto with zarith.
constructor.

intros z1 z2 l0 H0 H1.
elim (Z_lt_ge_dec x z2); elim (Z_lt_ge_dec x z1).

intros.
constructor. omega. (* auto with zarith.*)
constructor. omega.  
assumption.

intros.
constructor. omega.
assumption.

intros.
constructor. omega.
constructor. omega. 
assumption.

intros.
constructor. omega.
assumption.
Qed.



Theorem isort_correct : forall (l l':list Z),
                   l'=isort l -> Perm l l' /\ Sorted l'.
Proof.
induction l; intros.
unfold Perm; rewrite H; split; auto.
simpl.
constructor.
simpl in H.
rewrite H. (* ??????????? *)

elim (IHl (isort l)); intros; split.
apply Perm_trans with (a::isort l).
unfold Perm. intro z. simpl. elim (Z_eq_dec z a). intros.
elim H0; reflexivity. (* auto with zarith. *)
intros. elim H0. reflexivity.
Check insert_Perm.
apply insert_Perm.
Check insert_Sorted.
apply insert_Sorted.
assumption.
Qed.


(* EXTRACTION *)

(* Exercise: complete the following proof and extract its computational content to an Haskell 
function. 

Definition inssort : forall (l:list Z), { l' | Perm l l' & Sorted l' }.
induction l.

 ...

Defined.
*)

Extraction Language Haskell.

Extraction Inline list_rec.
Extraction Inline list_rect.
Extraction Inline sig2_rec.
Extraction Inline sig2_rect.

Extraction inssort.


(* 
 Exercise: use the same method to extract the insert function from its specification.
*)




(* ============================================================== *)

Close Scope Z_scope. 

Function div (p:nat*nat) {measure fst} : nat*nat :=
  match p with
  | (_,0) => (0,0)
  | (a,b) => if le_lt_dec b a
             then let (x,y):=div (a-b,b) in (1+x,y)
             else (0,a)
  end.
Proof.
intros.
simpl.
omega.
Qed.


(* Exercise: complete the definition of the function merge, presenting a proof of
its termination.

Function merge (p:list Z*list Z)
{measure (fun p=>(length (fst p))+(length (snd p)))} : list Z :=
  match p with
  | (nil,l) => l
  | (l,nil) => l
  | (x::xs,y::ys) => if Z_lt_ge_dec x y
                     then x::(merge (xs,y::ys))
                     else y::(merge (x::xs,ys))
  end.

*)



Definition divRel (args:nat*nat) (res:nat*nat) : Prop := 
 let (n,d):=args in let (q,r):=res in q*d+r=n /\ r<d. 

Definition divPre (args:nat*nat) : Prop := (snd args)<>0.


Theorem div_correct : forall (p:nat*nat),  divPre p -> divRel p (div p). 
Proof. 
unfold divPre, divRel. 
intro p. 
(* we make use of the specialised induction principle to conduct the proof... *) 
functional induction (div p); simpl. 
intro H; elim H; reflexivity. 
(* a first trick: we expand (div (a-b,b)) in order to get rid of the let (q,r)=... *) 
replace (div (a-b,b)) with (fst (div (a-b,b)),snd (div (a-b,b))) in IHp0. 
simpl in *. 
intro H; elim (IHp0 H); intros. 
split. 
(* again a similar trick: we expand "x" and "y0" in order to use an hypothesis *) 
change (b + (fst (x,y0)) * b + (snd (x,y0)) = a). 
rewrite <- e1. 
omega. 
(* and again... *) 
change (snd (x,y0)<b); rewrite <- e1; assumption. 
symmetry; apply surjective_pairing. 
auto. 
Qed. 